@@ -3,15 +3,16 @@ import logging
 
 from asana.error import NotFoundError, InvalidTokenError, ForbiddenError
 from django.apps import apps
-from django.db import IntegrityError, transaction
 from django.core.management.base import BaseCommand, CommandError
 from django.utils import six
 
 from djasana.connect import client_connect
 from djasana.models import (
-    Attachment, Project, Story, SyncToken, Tag, Task, Team, User, Webhook, Workspace)
+    Attachment, Project, Story, SyncToken,
+    Tag, Task, Team, User, Webhook, Workspace)
 from djasana.settings import settings
-from djasana.utils import set_webhook
+from djasana.utils import (
+    pop_unsupported_fields, set_webhook, sync_attachment, sync_project, sync_story, sync_task)
 
 logger = logging.getLogger(__name__)
 
@@ -22,6 +23,7 @@ class Command(BaseCommand):
     commit = True
     client = None
     process_archived = False
+    synced_ids = []  # A running list of remote ids of tasks that have been synced.
 
     @staticmethod
     def get_client():
@@ -47,6 +49,10 @@ class Command(BaseCommand):
             help='Sync only the named model (can be used multiple times). '
                  'By default all models will be updated from Asana.'
         )
+        parser.add_argument(
+            "-mx", "--model-exclude", action="append", default=[],
+            help='Exclude the named model (can be used multiple times).'
+        )
         parser.add_argument(
             "-a", "--archive", action="store_false", dest='archive',
             help='Sync project tasks etc. even if the project is archived. '
@@ -93,6 +99,7 @@ class Command(BaseCommand):
     def _get_models(options):
         """Returns a list of models to sync"""
         models = options.get('model')
+        models_exclude = options.get('model_exclude')
         app_models = list(apps.get_app_config('djasana').get_models())
         if models:
             good_models = []
@@ -107,6 +114,10 @@ class Command(BaseCommand):
             models = good_models
         else:
             models = app_models
+        if models_exclude:
+            models = [model
+                      for model in models
+                      if model.__name__.lower() not in [m.lower() for m in models_exclude]]
         return models
 
     def _check_sync_project_id(self, project_id, workspace, models):
@@ -116,7 +127,7 @@ class Command(BaseCommand):
             sync_token = SyncToken.objects.get(project_id=project_id)
             try:
                 events = self.client.events.get({'resource': project_id, 'sync': sync_token.sync})
-                self._process_events(project_id, events, workspace, models)
+                self._process_events(project_id, events, models)
                 self._set_webhook(workspace, project_id)
                 return
             except InvalidTokenError as error:
@@ -127,7 +138,7 @@ class Command(BaseCommand):
                 self.client.events.get({'resource': project_id})
             except InvalidTokenError as error:
                 new_sync = error.sync
-        is_archived = self._sync_project_id(project_id, workspace, models)
+        is_archived = self._sync_project_id(project_id, models)
         if not is_archived:
             self._set_webhook(workspace, project_id)
         if new_sync:
@@ -140,20 +151,20 @@ class Command(BaseCommand):
         if workspaces:
             for workspace in workspaces:
                 for wks in workspaces_:
-                    if workspace == str(wks['id']) or workspace == wks['name']:
-                        workspace_ids.append(wks['id'])
+                    if workspace in (wks['gid'], wks['name']):
+                        workspace_ids.append(wks['gid'])
                         break
                 else:
                     bad_list.append(workspace)
         else:
-            workspace_ids = [wks['id'] for wks in workspaces_]
+            workspace_ids = [wks['gid'] for wks in workspaces_]
         if bad_list:
             if len(bad_list) == 1:
                 raise CommandError('{} is not an Asana workspace'.format(workspaces[0]))
-            else:
-                raise CommandError('Specified workspaces are not valid: {}'.format(
-                    ', '.join(bad_list)))
-        return workspace_ids
+            raise CommandError('Specified workspaces are not valid: {}'.format(
+                ', '.join(bad_list)))
+        # Return newer workspaces first so they get synced earlier
+        return sorted(workspace_ids, reverse=True)
 
     def _get_project_ids(self, projects, workspace_id):
         project_ids = []
@@ -163,139 +174,135 @@ class Command(BaseCommand):
         if projects:
             for project in projects:
                 for prj in projects_:
-                    if project == str(prj['id']) or project == prj['name']:
-                        project_ids.append(prj['id'])
+                    if project in (prj['gid'], prj['name']):
+                        project_ids.append(prj['gid'])
                         break
                 else:
                     bad_list.append(project)
         else:
-            project_ids = [prj['id'] for prj in projects_]
+            project_ids = [prj['gid'] for prj in projects_]
         if bad_list:
             if len(bad_list) == 1:
                 raise CommandError('{} is not an Asana project'.format(bad_list[0]))
-            else:
-                raise CommandError('Specified projects are not valid: {}'.format(
-                    ', '.join(bad_list)))
-        return project_ids
+            raise CommandError('Specified projects are not valid: {}'.format(', '.join(bad_list)))
+        # Return newer projects first so they get synced earlier
+        return sorted(project_ids, reverse=True)
 
     def _set_webhook(self, workspace, project_id):
         """Sets a webhook if the setting is configured and a webhook does not currently exist"""
-        if self.commit and settings.DJASANA_WEBHOOK_URL:
-            webhooks = [webhook for webhook in self.client.webhooks.get_all({
-                'workspace': workspace.remote_id, 'resource': project_id})]
-            if webhooks:
-                # If there is exactly one, and it is active, we are good to go,
-                # else delete them and start a new one.
-                webhooks_ = Webhook.objects.filter(project_id=project_id)
-                if len(webhooks) == webhooks_.count() == 1:
-                    if webhooks[0]['active']:
-                        return
-                for webhook in webhooks:
-                    self.client.webhooks.delete_by_id(webhook['id'])
-                Webhook.objects.filter(id__in=webhooks_.values_list('id', flat=True)[1:]).delete()
-            set_webhook(self.client, project_id)
+        if not (self.commit and settings.DJASANA_WEBHOOK_URL):
+            return
+        webhooks = [webhook for webhook in self.client.webhooks.get_all({
+            'workspace': workspace.remote_id, 'resource': project_id})]
+        if webhooks:
+            # If there is exactly one, and it is active, we are good to go,
+            # else delete them and start a new one.
+            webhooks_ = Webhook.objects.filter(project_id=project_id)
+            if len(webhooks) == webhooks_.count() == 1 and webhooks[0]['active']:
+                return
+            for webhook in webhooks:
+                self.client.webhooks.delete_by_id(webhook['id'])
+            Webhook.objects.filter(id__in=webhooks_.values_list('id', flat=True)[1:]).delete()
+        set_webhook(self.client, project_id)
 
-    def _process_events(self, project_id, events, workspace, models):
+    def _process_events(self, project_id, events, models):
         project = Project.objects.get(remote_id=project_id)
+        ignored_tasks = 0
         for event in events['data']:
             if event['type'] == 'project':
-                if event['action'] == 'removed':
-                    Project.objects.get(remote_id=event['resource']['id']).delete()
+                if Project in models:
+                    if event['action'] == 'removed':
+                        Project.objects.get(remote_id=event['resource']['gid']).delete()
+                    else:
+                        self._sync_project_id(project_id, models)
                 else:
-                    self._sync_project_id(project_id, workspace, models)
+                    ignored_tasks += 1
             elif event['type'] == 'task':
-                if event['action'] == 'removed':
-                    Task.objects.get(remote_id=event['resource']['id']).delete()
+                if Task in models:
+                    if event['action'] == 'removed':
+                        Task.objects.get(remote_id=event['resource']['gid']).delete()
+                    else:
+                        self._sync_task(event['resource'], project, models)
                 else:
-                    self._sync_task(event['resource'], project, models)
+                    ignored_tasks += 1
             elif event['type'] == 'story':
-                self._sync_story(event['resource'])
-        message = 'Successfully synced {0} events for project {1}.'.format(
-            len(events['data']), project.name)
-        self.stdout.write(self.style.SUCCESS(message))
-        logger.info(message)
+                if Story in models:
+                    self._sync_story(event['resource'])
+                else:
+                    ignored_tasks += 1
+        tasks_done = len(events['data']) - ignored_tasks
+        if self.commit:
+            message = 'Successfully synced {0} events for project {1}.'.format(
+                tasks_done, project.name)
+            if ignored_tasks:
+                message += ' {0} events ignored for excluded models.'.format(ignored_tasks)
+            self.stdout.write(self.style.SUCCESS(message))
+            logger.info(message)
 
-    def _sync_project_id(self, project_id, workspace, models):
+    def _sync_project_id(self, project_id, models):
         """Sync this project by polling it. Returns boolean 'is archived?'"""
         project_dict = self.client.projects.find_by_id(project_id)
         logger.debug('Sync project %s', project_dict['name'])
         logger.debug(project_dict)
         if self.commit:
-            remote_id = project_dict.pop('id')
-            if project_dict['owner']:
-                owner = project_dict.pop('owner')
-                User.objects.get_or_create(remote_id=owner['id'], defaults={'name': owner['name']})
-                project_dict['owner_id'] = owner['id']
-            team = project_dict.pop('team')
-            Team.objects.get_or_create(remote_id=team['id'], defaults={'name': team['name']})
-            project_dict['team_id'] = team['id']
-            project_dict['workspace'] = workspace
-            # Convert string to boolean:
-            project_dict['archived'] = project_dict['archived'] == 'true'
-            members_dict = project_dict.pop('members')
-            followers_dict = project_dict.pop('followers')
-            project = Project.objects.update_or_create(
-                remote_id=remote_id, defaults=project_dict)[0]
-            member_ids = [member['id'] for member in members_dict]
-            members = User.objects.filter(id__in=member_ids)
-            project.members.set(members)
-            follower_ids = [follower['id'] for follower in followers_dict]
-            followers = User.objects.filter(id__in=follower_ids)
-            project.followers.set(followers)
-        else:
-            project = None
+            project = sync_project(self.client, project_dict)
 
-        if not project_dict['archived'] or self.process_archived:
+        if Task in models and not project_dict['archived'] or self.process_archived:
             for task in self.client.tasks.find_all({'project': project_id}):
                 self._sync_task(task, project, models)
-
-        if project:
+            # Delete local tasks for this project that are no longer in Asana.
+            tasks_to_delete = Task.objects.filter(projects=project).exclude(
+                remote_id__in=self.synced_ids).exclude(remote_id__isnull=True)
+            if tasks_to_delete.count() > 0:
+                id_list = list(tasks_to_delete.values_list('remote_id', flat=True))
+                tasks_to_delete.delete()
+                message = 'Deleted {} tasks no longer present: {}'.format(len(id_list), id_list)
+                self.stdout.write(self.style.SUCCESS(message))
+                logger.info(message)
+        if self.commit:
             message = 'Successfully synced project {}.'.format(project.name)
             self.stdout.write(self.style.SUCCESS(message))
             logger.info(message)
         return project_dict['archived']
 
     def _sync_story(self, story):
+        story_id = story.get('gid')
         try:
-            story_dict = self.client.stories.find_by_id(story['id'])
+            story_dict = self.client.stories.find_by_id(story_id)
         except NotFoundError as error:
             logger.info(error.response)
             return
         logger.debug(story_dict)
-        remote_id = story_dict.pop('id')
-        if story_dict['created_by']:
-            user = User.objects.get_or_create(
-                remote_id=story_dict['created_by']['id'],
-                defaults={'name': story_dict['created_by']['name']})[0]
-            story_dict['created_by'] = user
-        if story_dict['target']:
-            story_dict['target'] = story_dict['target']['id']
-        for key in ('hearts', 'liked', 'likes', 'num_likes'):
-            story_dict.pop(key, None)
-        if 'text' in story_dict:
-            story_dict['text'] = story_dict['text'][:1024]  # Truncate text if too long
-        Story.objects.get_or_create(remote_id=remote_id, defaults=story_dict)
+        remote_id = story_dict['gid']
+        sync_story(remote_id, story_dict)
 
-    def _sync_tag(self, tag):
-        tag_dict = self.client.tags.find_by_id(tag['id'])
+    def _sync_tag(self, tag, workspace):
+        tag_dict = self.client.tags.find_by_id(tag['gid'])
         logger.debug(tag_dict)
         if self.commit:
-            remote_id = tag_dict.pop('id')
-            Tag.objects.get_or_create(
+            remote_id = tag_dict['gid']
+            tag_dict['workspace'] = workspace
+            followers_dict = tag_dict.pop('followers')
+            pop_unsupported_fields(tag_dict, Tag)
+            tag = Tag.objects.get_or_create(
                 remote_id=remote_id,
-                defaults=tag_dict)
+                defaults=tag_dict)[0]
+            follower_ids = [follower['gid'] for follower in followers_dict]
+            followers = User.objects.filter(id__in=follower_ids)
+            tag.followers.set(followers)
 
     def _sync_task(self, task, project, models, skip_subtasks=False):
-        """Sync this task and parent its subtasks
+        """Sync this task and its parent, dependencies, and subtasks
 
         For parents and subtasks, this method is called recursively, so skip_subtasks True is
         passed when syncing a parent task from a subtask.
         """
+        task_id = task['gid']
         try:
-            task_dict = self.client.tasks.find_by_id(task['id'])
+            task_dict = self.client.tasks.find_by_id(task_id)
         except (ForbiddenError, NotFoundError):
             try:
-                Task.objects.get(remote_id=task['id']).delete()
+                Task.objects.get(remote_id=task_id).delete()
             except Task.DoesNotExist:
                 pass
             return
@@ -303,69 +310,54 @@ class Command(BaseCommand):
         logger.debug(task_dict)
 
         if Task in models and self.commit:
-            remote_id = task_dict.pop('id')
-            if task_dict['assignee']:
-                user = User.objects.get_or_create(
-                    remote_id=task_dict['assignee']['id'],
-                    defaults={'name': task_dict['assignee']['name']})[0]
-                task_dict['assignee'] = user
-            for key in (
-                    'hearts', 'liked', 'likes', 'num_likes',
-                    'memberships', 'projects', 'workspace'):
-                task_dict.pop(key, None)
+            remote_id = task_dict['gid']
             parent = task_dict.pop('parent', None)
+            dependencies = task_dict.pop('dependencies', None) or []
             if parent:
                 # If this is a task we already know about, assume it was just synced.
-                parent_id = parent['id']
-                if not Task.objects.filter(remote_id=parent_id).exists():
+                parent_id = parent['gid']
+                if parent_id not in self.synced_ids and \
+                        not Task.objects.filter(remote_id=parent_id).exists():
                     self._sync_task(parent, project, models, skip_subtasks=True)
                 task_dict['parent_id'] = parent_id
-            followers_dict = task_dict.pop('followers')
-            tags_dict = task_dict.pop('tags')
-            task_ = Task.objects.update_or_create(
-                remote_id=remote_id, defaults=task_dict)[0]
+            task_ = sync_task(remote_id, task_dict, project, sync_tags=Tag in models)
+            self.synced_ids.append(remote_id)
             if not skip_subtasks:
-                subtasks = self.client.tasks.subtasks(task['id'])
-                for subtask in subtasks:
-                    self._sync_task(subtask, project, models)
-            follower_ids = [follower['id'] for follower in followers_dict]
-            followers = User.objects.filter(id__in=follower_ids)
-            task_.followers.set(followers)
-            for tag_ in tags_dict:
-                tag = Tag.objects.get_or_create(
-                    remote_id=tag_['id'],
-                    defaults={'name': tag_['name']})[0]
-                task_.tags.add(tag)
-            task_.projects.add(project)
+                for subtask in self.client.tasks.subtasks(task_id):
+                    if subtask['gid'] not in self.synced_ids:
+                        self._sync_task(subtask, project, models)
+                if dependencies:
+                    for subtask in dependencies:
+                        if subtask['gid'] not in self.synced_ids:
+                            self._sync_task(subtask, project, models)
+                    task_.dependencies.set(
+                        Task.objects.filter(remote_id__in=[dep['gid'] for dep in dependencies]))
         if Attachment in models and self.commit:
-            for attachment in self.client.attachments.find_by_task(task['id']):
-                attachment_dict = self.client.attachments.find_by_id(attachment['id'])
-                logger.debug(attachment_dict)
-                remote_id = attachment_dict.pop('id')
-                if attachment_dict['parent']:
-                    attachment_dict['parent'] = task_
-                Attachment.objects.get_or_create(remote_id=remote_id, defaults=attachment_dict)
+            for attachment in self.client.attachments.find_by_task(task_id):
+                sync_attachment(self.client, task_, attachment['gid'])
         if Story in models and self.commit:
-            for story in self.client.stories.find_by_task(task['id']):
+            for story in self.client.stories.find_by_task(task_id):
                 self._sync_story(story)
+        return
 
     def _sync_team(self, team):
-        team_dict = self.client.teams.find_by_id(team['id'])
+        team_dict = self.client.teams.find_by_id(team['gid'])
         logger.debug(team_dict)
         if self.commit:
-            remote_id = team_dict.pop('id')
+            remote_id = team_dict['gid']
             organization = team_dict.pop('organization')
-            team_dict['organization_id'] = organization['id']
+            team_dict['organization_id'] = organization['gid']
             team_dict['organization_name'] = organization['name']
+            pop_unsupported_fields(team_dict, Team)
             Team.objects.get_or_create(
                 remote_id=remote_id,
                 defaults=team_dict)
 
     def _sync_user(self, user, workspace):
-        user_dict = self.client.users.find_by_id(user['id'])
+        user_dict = self.client.users.find_by_id(user['gid'])
         logger.debug(user_dict)
         if self.commit:
-            remote_id = user_dict.pop('id')
+            remote_id = user_dict['gid']
             user_dict.pop('workspaces')
             if user_dict['photo']:
                 user_dict['photo'] = user_dict['photo']['image_128x128']
@@ -380,7 +372,7 @@ class Command(BaseCommand):
         logger.debug('Sync workspace %s', workspace_dict['name'])
         logger.debug(workspace_dict)
         if Workspace in models and self.commit:
-            remote_id = workspace_dict.pop('id')
+            remote_id = workspace_dict['gid']
             workspace_dict.pop('email_domains')
             workspace = Workspace.objects.update_or_create(
                 remote_id=remote_id, defaults=workspace_dict)[0]
@@ -388,7 +380,7 @@ class Command(BaseCommand):
             workspace = None
         project_ids = self._get_project_ids(projects, workspace_id)
         if workspace_id != self.client.options['workspace_id']:
-            self.client.options['workspace_id'] = workspace_id
+            self.client.options['workspace_id'] = str(workspace_id)
 
         if User in models:
             for user in self.client.users.find_all({'workspace': workspace_id}):
@@ -396,7 +388,7 @@ class Command(BaseCommand):
 
         if Tag in models:
             for tag in self.client.tags.find_by_workspace(workspace_id):
-                self._sync_tag(tag)
+                self._sync_tag(tag, workspace)
 
         if Team in models:
             for team in self.client.teams.find_by_organization(workspace_id):
