@@ -6,17 +6,17 @@ from django.db import IntegrityError
 from django.test import override_settings, TestCase
 from djasana.management.commands.sync_from_asana import Command
 from djasana.models import (
-    Attachment, Project, Story, SyncToken, Tag, Task, Team, Webhook, Workspace, User)
+    Attachment, CustomField, CustomFieldSetting, Project, Story,
+    SyncToken, Tag, Task, Team, Webhook, Workspace, User)
 from djasana.tests.fixtures import (
-    attachment, project, story, tag, task, team, user, webhook, workspace)
+    attachment, custom_field, project, story, tag, task, team, user, webhook, workspace)
 
 
 def mock_connect():
     return Mock()
 
 
-@override_settings(ASANA_ACCESS_TOKEN='foo')
-@override_settings(ASANA_WORKSPACE=None)
+@override_settings(ASANA_ACCESS_TOKEN='foo', ASANA_WORKSPACE=None)
 class CommandArgumentsTestCase(TestCase):
     """Tests of command argument handling that do not call Asana
     (adds coverage to add_arguments)
@@ -28,10 +28,30 @@ class CommandArgumentsTestCase(TestCase):
         with self.assertRaises(CommandError):
             call_command('sync_from_asana', *args, **options)
 
+    @patch('djasana.management.commands.sync_from_asana.Command._sync_workspace_id')
+    @patch('djasana.management.commands.sync_from_asana.Command._get_workspace_ids')
+    def test_models(self, mock_workspace_ids, mock_sync):
+        mock_workspace_ids.return_value = [1]
+        args = ['--noinput']
+        options = {'model': ['Workspace']}
+        call_command('sync_from_asana', *args, **options)
+        self.assertEqual(1, len(mock_sync.call_args[0][2]))
+        self.assertTrue(Workspace in mock_sync.call_args[0][2])
 
-@override_settings(ASANA_ACCESS_TOKEN='foo')
-@override_settings(ASANA_WORKSPACE=None)
-@override_settings(ROOT_URLCONF='djasana.urls')
+    @patch('djasana.management.commands.sync_from_asana.Command._sync_workspace_id')
+    @patch('djasana.management.commands.sync_from_asana.Command._get_workspace_ids')
+    def test_models_exclude(self, mock_workspace_ids, mock_sync):
+        mock_workspace_ids.return_value = [1]
+        args = ['--noinput']
+        models_to_exclude = ['Story', 'Tags', 'Attachment']
+        options = {'model_exclude': models_to_exclude}
+        call_command('sync_from_asana', *args, **options)
+        self.assertTrue(Workspace in mock_sync.call_args[0][2])
+        for model in models_to_exclude:
+            self.assertFalse(model in mock_sync.call_args[0][2])
+
+
+@override_settings(ASANA_ACCESS_TOKEN='foo', ASANA_WORKSPACE=None, ROOT_URLCONF='djasana.urls')
 class SyncFromAsanaTestCase(TestCase):
     """Tests that use mock returns from Asana"""
 
@@ -65,7 +85,6 @@ class SyncFromAsanaTestCase(TestCase):
         self.assertEqual(1, Attachment.objects.count())
         self.assertEqual(1, Story.objects.count())
 
-    @override_settings(ASANA_WORKSPACE=None)
     def test_good_workspace(self):
         self.command.handle(interactive=False, workspace=['Test Workspace'])
         self.assertEqual(1, Workspace.objects.count())
@@ -140,33 +159,33 @@ class SyncFromAsanaTestCase(TestCase):
                     'created_at': '2017-08-21T18:20:37.972Z',
                     'parent': None,
                     'resource': {
-                        'id': 1,
+                        'gid': '1',
                         'name': 'Test Project'
                     },
                     'type': 'project',
-                    'user': 1123
+                    'user': '1123'
                 },
                 {
                     'action': 'changed',
                     'created_at': '2017-08-21T18:20:37.972Z',
                     'parent': None,
                     'resource': {
-                        'id': 1337,
+                        'gid': '1337',
                         'name': 'Test Task'
                     },
                     'type': 'task',
-                    'user': 1123
+                    'user': '1123'
                 },
                 {
                     'action': 'added',
                     'created_at': '2017-08-21T18:20:37.972Z',
                     'parent': None,
                     'resource': {
-                        'id': 1,
+                        'gid': '1',
                         'name': 'Test Story'
                     },
                     'type': 'story',
-                    'user': 1123
+                    'user': '1123'
                 }
             ]
         }
@@ -198,20 +217,48 @@ class SyncFromAsanaTestCase(TestCase):
         self.command.client.stories.find_by_task.return_value = [long_story]
         self.command.client.stories.find_by_id.return_value = long_story
         self.command.handle(interactive=False, verbosity=2)
-        self.assertTrue(1, Story.objects.exists())
+        self.assertTrue(Story.objects.exists())
         story_instance = Story.objects.last()
         self.assertEqual(1024, len(story_instance.text))
 
     def test_task_with_parent(self):
+        # When processed, tasks get modified in place; we need to pass the original twice.
         parent_task = task()
-        child_task = task(id=2, parent=parent_task)
-        self.command.client.tasks.find_all.return_value = [child_task]
-        self.command.client.tasks.find_by_id.side_effect = [child_task, parent_task]
+        parent_copy = parent_task.copy()
+        child_task = task(gid='2', parent=parent_task)
+        self.command.client.tasks.find_all.return_value = [child_task, parent_task]
+        self.command.client.tasks.find_by_id.side_effect = [child_task, parent_task, parent_copy]
         self.command.handle(interactive=False, project=['Test Project'])
         self.assertEqual(2, Task.objects.count())
         parent, child = tuple(Task.objects.order_by('remote_id'))
         self.assertEqual(parent, child.parent)
 
+    def test_subtask(self):
+        """Asserts subtask (task related to a task but not a project) is supported."""
+        parent_task = task(projects=None)  # Parent is also a subtask
+        subtask = task(gid='2', projects=None, parent=parent_task.copy())
+
+        self.command.client.tasks.find_all.return_value = [parent_task.copy()]
+        self.command.client.tasks.subtasks.side_effect = [[subtask.copy()], []]
+        self.command.client.tasks.find_by_id.side_effect = [parent_task.copy(), subtask.copy()]
+
+        self.command.handle(interactive=False, workspace=['Test Workspace'])
+        self.assertEqual(2, Task.objects.count())
+
+    def test_task_not_in_asana_is_deleted(self):
+        workspace_ = Workspace.objects.create(remote_id=1, name='Workspace')
+        team_ = Team.objects.create(remote_id=2, name='Team')
+        project_ = Project.objects.create(
+            remote_id=3, name='Another Project', public=True, team=team_, workspace=workspace_)
+        task_ = Task.objects.create(remote_id=4, name='Orphan Task', completed=False)
+        task_.projects.add(project_)
+        project_dict = project(gid='3')
+        self.command.client.projects.find_all.return_value = [project_dict]
+        self.command.client.projects.find_by_id.return_value = project_dict
+        self.command.client.tasks.find_all.return_value = []
+        self.command.handle(interactive=False, project=['Test Project'])
+        self.assertFalse(Task.objects.filter(pk=task_.pk).exists())
+
     @override_settings(DJASANA_WEBHOOK_URL='https://example.com/hooks/')
     def test_redundant_webhooks_are_deleted(self):
         workspace_ = Workspace.objects.create(remote_id=1, name='Workspace')
@@ -221,7 +268,7 @@ class SyncFromAsanaTestCase(TestCase):
         secret = 'x' * 32
         Webhook.objects.create(secret=secret, project=project_)
         Webhook.objects.create(secret=secret, project=project_)
-        project_dict = project(id=3)
+        project_dict = project(gid='3')
         self.command.client.projects.find_all.return_value = [project_dict]
         self.command.client.projects.find_by_id.return_value = project_dict
         webhook_ = webhook(project=project_dict)
@@ -231,8 +278,46 @@ class SyncFromAsanaTestCase(TestCase):
         self.assertEqual(1, Webhook.objects.filter(project=project_).count())
 
     def test_subtasks_synced(self):
-        child_task = task(id=99, name='Subtask', parent=task())
-        self.command.client.tasks.find_by_id.side_effect = [task(), child_task]
-        self.command.client.tasks.subtasks.side_effect = [[child_task], []]
+        parent_task = task()
+        child_task = task(gid='99', name='Subtask', parent=task())
+        # When processed, tasks get modified in place; we need to pass the original twice.
+        child_copy = child_task.copy()
+        self.command.client.tasks.find_all.return_value = [parent_task, child_task]
+        self.command.client.tasks.find_by_id.side_effect = [parent_task, child_task, child_copy]
+        self.command.client.tasks.subtasks.side_effect = [[child_task], [], []]
         self.command.handle(interactive=False)
         self.assertTrue(Task.objects.filter(remote_id=99, name='Subtask').exists())
+
+    def test_dependencies_synced(self):
+        main_task = task(gid='98', name='Is a dependency for 99', dependents=[task(gid='99')])
+        dependent_task = task(gid='99', name='Requires task 98', dependencies=[task(gid='98')])
+        # When processed, tasks get modified in place; we need to pass the original twice.
+        dependent_copy = dependent_task.copy(), dependent_task.copy()
+        self.command.client.tasks.find_all.return_value = [main_task, dependent_task]
+        self.command.client.tasks.find_by_id.side_effect = [
+            main_task, dependent_task, dependent_copy]
+        self.command.handle(interactive=False)
+        self.assertTrue(Task.objects.filter(remote_id=98).exists())
+        self.assertTrue(Task.objects.filter(remote_id=99).exists())
+        main_task = Task.objects.get(remote_id=98)
+        dependent_task = Task.objects.get(remote_id=99)
+        self.assertTrue(main_task in dependent_task.dependencies.all())
+        self.assertTrue(dependent_task in main_task.dependents.all())
+
+    def test_custom_fields(self):
+        workspace_ = Workspace.objects.create(remote_id=1, name='Workspace')
+        team_ = Team.objects.create(remote_id=2, name='Team')
+        Project.objects.create(
+            remote_id=3, name='New Project', public=True, team=team_, workspace=workspace_)
+        self.command.client.custom_fields.find_by_id.return_value = custom_field().copy()
+        self.command.client.projects.find_by_id.return_value = project(
+            gid='3',
+            custom_field_settings=[{
+                'gid': '258147',
+                'custom_field': custom_field().copy(),
+                'project': {'gid': '3', 'name': 'New Project'}}],
+            workspace={'gid': '1', 'name': 'Workspace'})
+        self.command._sync_project_id(project_id=3, models=[])
+        self.assertTrue(CustomField.objects.filter(remote_id=1, name='Test Custom Field').exists())
+        self.assertTrue(
+            CustomFieldSetting.objects.filter(remote_id=258147, project_id=3).exists())
